import logging
from time import time
import os
from sentry_sdk_extensions.nohub import capture_exception

from django.contrib.auth.decorators import login_required
from django.db import connection
from django.conf import settings
from django.core.exceptions import SuspiciousOperation

performance_logger = logging.getLogger("bugsink.performance.views")


class DisallowChunkedMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if "HTTP_TRANSFER_ENCODING" in request.META and \
                request.META["HTTP_TRANSFER_ENCODING"].lower() == "chunked" and \
                not request.META.get("wsgi.input_terminated"):

            # If we get here, it means that the request has a Transfer-Encoding header with a value of "chunked", but we
            # have no wsgi-layer handling for that. This probably means that we're running the Django development
            # server, and as such our fixes for the Gunicorn/Django mismatch that we put in wsgi.py are not in effect.
            raise ValueError("This server is not configured to support Chunked Transfer Encoding (for requests)")

        return self.get_response(request)


class LoginRequiredMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_view(self, request, view_func, view_args, view_kwargs):
        # returning None is interpreted by Django as "proceed with handling it"; we do this for all cases where we
        # have determined that the user is authenticated or user authentication is not required

        if request.user.is_authenticated:
            return None

        # we explicitly ignore the admin and accounts paths, and the api; we can always push this to a setting later
        for path in ["/admin", "/accounts", "/api"]:
            if request.path.startswith(path):
                return None

        if getattr(view_func, 'login_exempt', False):
            return None

        # Note: this short-circuits the rest of the middleware handling. That's OK in this case though, because what we
        # do amounts to "redirect and do something else". (It's not OK in the general case, where you just want to do a
        # small middleware-like thing and proceed normally otherwise)
        return login_required(view_func)(request, *view_args, **view_kwargs)


class PerformanceStatsMiddleware:
    """TSTTCPW to get some handle on view-performance (mostly for UI views). The direct cause for introducing this is
    that I got sent on a wild goose chase by the Django Debug Toolbar, which reported long (>100ms) CPU times for some
    view when it was the DJDT itself that was causing most of that time.
    """

    def __init__(self, get_response):
        self.get_response = get_response
        self.view_name = "<<unknown>>"

    def __call__(self, request):
        t0 = time()
        result = self.get_response(request)
        took = (time() - t0) * 1000
        performance_logger.info(f"{took:6.2f}ms / {len(connection.queries)} queries: '{ self.view_name }'")
        return result

    def process_view(self, request, view_func, view_args, view_kwargs):
        self.view_name = view_func.__name__
        return None


class CaptureExceptionsMiddleware:
    # Capture exceptions using a Middleware rather than the more magical sentry_sdk.init() method. This is more
    # explicit and easier to understand, though less feature-complete.
    # Debugging tool (for development); not turned on by default.

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_exception(self, request, exception):
        dsn = os.getenv("SENTRY_DSN")
        if dsn is not None:
            capture_exception(dsn, exception)
        return None


class SetRemoteAddrMiddleware:
    """
    Sets the REMOTE_ADDR from the proxy headers if so configured. Sets REMOTE_ADDR to None if the configured headers are
    empty (misconfiguration), i.e. None rather than 127.0.0.1 in that case.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    @staticmethod
    def parse_x_forwarded_for(header_value):
        if header_value in [None, ""]:
            # The most typical misconfiguration is to forget to set the header at all, or to have it be empty. In that
            # case, we'll just set the IP to None, which will mean some data will be missing from your events (but
            # you'll still get them).
            return None

        # Elements are comma-separated, with optional whitespace surrounding the commas. (MDN)
        ips = [s.strip() for s in header_value.split(",")]

        # Each proxy in the chain appends the IP it is forwarding for to the list. Hence, the number of proxies you use
        # must exactly equal the number of IPs in the list. (+1 for the client; -1 because the last proxy will not be in
        # the list)
        if len(ips) == settings.X_FORWARDED_FOR_PROXY_COUNT:
            return ips[0]  # The first address is the original client; others are proxies

        if len(ips) > settings.X_FORWARDED_FOR_PROXY_COUNT:
            # Greater than: somebody added something, most likely maliciously. Complain loudly.
            # Alternatively, we could just take the one we trust (at index -proxy_count), but if someone's spoofing
            # there's no reason to trust the rest of the message (event data).
            raise SuspiciousOperation("X-Forwarded-For header does not contain the expected number of addresses")

        # implied: len(ips) < settings.X_FORWARDED_FOR_PROXY_COUNT:
        # As in 'if header_value is None' above, this is a misconfiguration. We'll just set the IP to None.
        return None

    def __call__(self, request):
        if settings.USE_X_REAL_IP:
            request.META["REMOTE_ADDR"] = request.META.get("HTTP_X_REAL_IP", None)

        elif settings.USE_X_FORWARDED_FOR:  # elif: X-Real-IP / X-Forwarded-For are mutually exclusive
            request.META["REMOTE_ADDR"] = self.parse_x_forwarded_for(request.META.get("HTTP_X_FORWARDED_FOR", None))

        return self.get_response(request)



# Generated by Django 4.2.16 on 2024-11-20 13:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("teams", "0002_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="team",
            name="visibility",
            field=models.IntegerField(
                choices=[(1, "Joinable"), (10, "Discoverable"), (99, "Hidden")],
                default=10,
                help_text="Which users can see this team and its issues?",
            ),
        ),
    ]



from urllib.parse import urlparse

from django.core.mail import EmailMultiAlternatives
from django.template.loader import get_template

import sentry_sdk
from sentry_sdk_extensions.transport import MoreLoudlyFailingTransport


def send_rendered_email(subject, base_template_name, recipient_list, context=None):
    if context is None:
        context = {}

    html_content = get_template(base_template_name + ".html").render(context)
    text_content = get_template(base_template_name + ".txt").render(context)

    # Configure and send an EmailMultiAlternatives
    msg = EmailMultiAlternatives(
        subject=subject,
        body=text_content,
        from_email=None,  # this is settings.DEFAULT_FROM_EMAIL
        to=recipient_list,
    )

    msg.attach_alternative(html_content, "text/html")

    msg.send()


def deduce_allowed_hosts(base_url):
    url = urlparse(base_url)
    if url.hostname == "localhost" or url.hostname == "127.0.0.1":
        # Allow both 127.0.0.1 and localhost as hostname when running locally; useful to avoid understanding why
        # localhost:8000 is not allowed when 127.0.0.1:8000 is and vice versa.
        return ["localhost", "127.0.0.1"]
    return [url.hostname]


def eat_your_own_dogfood(sentry_dsn, **kwargs):
    """
    Configures your Bugsink installation to send messages to some Bugsink-compatible installation.
    See https://www.bugsink.com/docs/dogfooding/
    """

    if sentry_dsn is None:
        return

    default_kwargs = {
        "dsn": sentry_dsn,
        "traces_sample_rate": 0,
        "send_default_pii": True,
        "transport": MoreLoudlyFailingTransport,

        # see (e.g.) https://github.com/getsentry/sentry-python/issues/377 for why this is necessary; I really really
        # dislike Sentry's silent dropping of local variables; let's see whether "just send everything" makes for
        # messages that are too big. If so, we might monkey-patch sentry_sdk/serializer.py 's 2 variables named
        # MAX_DATABAG_DEPTH and MAX_DATABAG_BREADTH (esp. the latter)
        "max_request_body_size": "always",

        # In actual development, the list below is not needed, because in that case Sentry's SDK is able to distinguish
        # based on the os.cwd() v.s. site-packages. For cases where the Production installation instructions are
        # followed, that doesn't fly though, because we "just install everything" (using pip install), and we need to be
        # explicit. The notation below (no trailing dot or slash) is the correct one (despite not being documented) as
        # evidenced by the line `if item == name or name.startswith(item + "."):` in the sentry_sdk source:
        "in_app_include": [
            "alerts",
            "bugsink",
            "compat",
            "events",
            "ingest",
            "issues",
            "performance",
            "projects",
            "releases",
            "sentry",
            "sentry_sdk_extensions",
            "snappea",
            "teams",
            "theme",
            "users",
        ],
    }

    default_kwargs.update(kwargs)

    sentry_sdk.init(
        **default_kwargs,
    )



from datetime import timedelta

from django.utils import timezone
from django.utils.safestring import mark_safe

from bugsink.app_settings import get_settings, CB_ANYBODY
from bugsink.transaction import durable_atomic

from snappea.settings import get_settings as get_snappea_settings
from snappea.models import Task

from phonehome.models import Installation


FREE_VERSION_WARNING = mark_safe(
    """This is the free version of Bugsink; usage is limited to a single user for local development only.
    Using this software in production requires a
    <a href="https://www.bugsink.com/#pricing" target="_blank" class="font-bold text-slate-800">paid licence</a>.""")


def get_snappea_warnings():
    # We warn in either of 2 cases, as documented per-case.

    if get_snappea_settings().TASK_ALWAYS_EAGER:
        # In TASK_ALWAYS_EAGER mode, there is no expectation that snappea runs, so there should be no warning either.
        return []

    with durable_atomic(using="snappea"):
        task_count = Task.objects.all().count()
        if task_count == 0:
            # No tasks, no warnings.
            return []

        oldest_task_age = (timezone.now() - Task.objects.all().order_by('created_at').first().created_at).seconds

    WARNING = (f"Snappea has {task_count} tasks in the queue, the oldest being {oldest_task_age}s old. It may be "
               f"either overwhelmed, blocked, not running, or misconfigured.")

    # 1. "a lot" of tasks in the backlog.
    # We have a backlog because spikes are always to be expected, and dealing with them in a backlog is the feature that
    # makes the server reliably able to handle them. So why would we warn about the backlog actually being used? Well,
    # if there are "enough" (A.K.A. "a lot") of items in the backlog, you might be confused "why isn't this showing up?"
    # 10s of confusion is a nice cut-off after which a banner should unconfuse you, and with 30 events per second (safe
    # assumption given well-documented measurements) this is 300 events. Also: assumed to be rare enough to not
    # over-warn.
    if task_count > 300:
        return [WARNING]

    # 2. "a few", but _stale_ tasks in the backlog.
    # If there are some (not a lot, but >0) items in the backlog, but they seem "stale", we warn too. Hard to come up
    # with a precise definition of "stale", but we can say that if the oldest task is older than 5 seconds, people may
    # start wondering why it's not happening. (whether that is being caused by backlogging or misconfiguration, in
    # either case the warning is useful)". (An even better measure for low task_counts would be: if the last _picked-up_
    # task is longer than some time ago, but we don't register picked-up time because we delete tasks in that case). For
    # now, that's not complexity I'm willing to introduce, so we stick with the arbitrary 5s threshold.
    if oldest_task_age > 5:
        return [WARNING]

    return []


def useful_settings_processor(request):
    # name is misnomer, but "who cares".

    installation = Installation.objects.get()

    nag_7 = installation.created_at < timezone.now() - timedelta(days=7)
    nag_30 = installation.created_at < timezone.now() - timedelta(days=30)

    # (First version of "should I nag" logic): nag only after considerable time to play with the app, and for "some
    # indication" that you're using this in production (the simplest such indication is that you've configured a
    # BASE_URL that's not localhost). Subject to change.
    system_warnings = [FREE_VERSION_WARNING] if nag_30 and 'localhost' not in get_settings().BASE_URL else []

    return {
        # Note: no way to actually set the license key yet, so nagging always happens for now.
        'site_title': get_settings().SITE_TITLE + (" (non-production use)" if nag_7 else ""),
        'registration_enabled': get_settings().USER_REGISTRATION == CB_ANYBODY,
        'app_settings': get_settings(),
        'system_warnings': system_warnings + get_snappea_warnings(),
    }


def logged_in_user_processor(request):
    return {
        'logged_in_user': request.user,
    }



import logging
from user_agents import parse as ua_parse

logger = logging.getLogger("bugsink.events.ua_stuff")


def enrich_contexts_with_ua(parsed_data):
    # GlitchTip has some mechanism to get "synthetic" (i.e. not present in the original, UA-header derived) info into
    # first the contexts, which is then propagated (with a whole bunch of other info from contexts) to the tags. Both
    # these steps happen on-digest.
    #
    # I'm not sure what I want myself... on the one hand I'm not a fan of this kind of magic. On the other: if the data
    # usually lives in the "context" (e.g. in the JS world), you might as well synthesize it into that location when it
    # is not in that location in the data (but is available through other means).
    #
    # My set of samples has very little data for contexts, so it's hard to get a feel for it. I imagine that having this
    # info available in tags can be useful for searching, or for getting a quick feel of the data (tags show up in the
    # RHS of various screens). But again: too little data to tell yet. Add to that that I'm much less inclined than my
    # competitors to give OS/browser info the main stage (icons? yuck!). So we'll just parse it, put it "somewhere", and
    # look at it again "later".

    try:
        contexts = parsed_data.get("contexts", {})

        ua_string = (parsed_data.get("request", {}).get("headers", {})).get("User-Agent")
        if ua_string is None:
            return contexts

        if isinstance(ua_string, list):
            if len(ua_string) == 0:
                return contexts
            ua_string = ua_string[0]  # assuming: it's always just one, and if it's not we just pick that anyway

        user_agent = ua_parse(ua_string)

        if "browser" not in contexts:
            contexts["browser"] = {
                "name": user_agent.browser.family,
                "version": user_agent.browser.version_string,
            }

        if "os" not in contexts:
            contexts["os"] = {
                "name": user_agent.os.family,
                "version": user_agent.os.version_string,
            }

        if "device" not in contexts:
            contexts["device"] = {
                "family": user_agent.device.family,
                "model": user_agent.device.model,
                "brand": user_agent.device.brand,
            }
    except Exception as e:
        # We take the approach of "do not fail to display the event" here. If we can't enrich the contexts with UA info,
        # we'll just log it and move on.

        logger.warning(
            "Failed to enrich contexts with User-Agent info for %s (%s)",
            ua_string, type(ua_string).__name__, exc_info=e)

    return contexts



from django.test import TransactionTestCase


from phonehome.models import Installation


class TransactionTestCase25251(TransactionTestCase):
    """
    As per https://code.djangoproject.com/ticket/25251 "Any initial data loaded in migrations will only be available
    in TestCase tests and not in TransactionTestCase tests", see also:
    https://docs.djangoproject.com/en/5.1/topics/testing/overview/#rollback-emulation

    i.e. a documented footgun. And the hardest type to find, b/c the tests _do_ work in isolation (but not in batches).

    We work around this by doing the work manually here. As discovered by observing crashing tests, and by
    $ git ls-files | grep migrat | x grep object.*create
    """

    def setUp(self):
        super().setUp()
        Installation.objects.get_or_create()  # get_or_create, b/c on the first run the migration _is_ there.



from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter

from django import template


register = template.Library()


@register.tag(name="code")
def do_code(parser, token):
    nodelist = parser.parse(("endcode",))
    parser.delete_first_token()
    return CodeNode(nodelist)


class CodeNode(template.Node):

    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        content = self.nodelist.render(context)

        # remove trailing whitespace from each line (it's never wanted, and in the case of code blocks it can actually
        # hurt, e.g. for space-after-backslash in bash)
        content = "\n".join([line.rstrip() for line in content.split("\n")])

        lang_identifier, code = content.split("\n", 1)
        assert lang_identifier.startswith(":::") or lang_identifier.startswith("#!"), \
            "Expected code block identifier ':::' or '#!' not " + lang_identifier

        lang = lang_identifier[3:].strip() if lang_identifier.startswith(":::") else lang_identifier[2:].strip()
        is_shebang = lang_identifier.startswith("#!")
        formatter = HtmlFormatter(linenos="table" if is_shebang else False)

        lexer = get_lexer_by_name(lang, stripall=True)

        return highlight(code, lexer, formatter).replace("highlight", "p-4 mt-4 bg-slate-50 syntax-coloring")



# Generated by Django 4.2.16 on 2024-11-11 12:48

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ("phonehome", "0003_outboundmessage"),
    ]

    operations = [
        migrations.AddField(
            model_name="installation",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
    ]
